from epure import epure
from ..tg_environment.t_user import TUser
from ..finance.currency import Currency
from ..tg_environment.t_message import TMessage
from .offer_state import OfferState
from .offer_type import OfferType
from uuid import UUID
from datetime import datetime

@epure()
class TOffer:
    from_user: TUser
    to_user: TUser
    offer_type: OfferType
    state: OfferState
    subject: TMessage
    currency: Currency
    price: float
    time: datetime
    duration: int
    message: TMessage
    bidding_id: UUID


from aiogram import Bot, Dispatcher, types, Router
from .tg_environment.t_user import TUser
from .taalc_bot import TaalcBot
from re import Match
import traceback
from aiogram.types import ChatMemberUpdated
from aiogram.types import Message, Update
from .teest.testing_message import TestingMessage

async def handle_with_tests(handler, message, match=None):
    try:
        user = TUser.user_by_tg_user(message.from_user)
        if TaalcBot.testers:
            message = TestingMessage(
                text = message.text,
                message_id = message.message_id,
                date = message.date,
                chat = message.chat,
                message_thread_id = message.message_thread_id,
                from_user = message.from_user,
                via_bot = message.via_bot,
                reply_to_message=message.reply_to_message
            ).as_(message.bot)
        
        if match:
            result =  await handler(message, user, match)
        else:
            result =  await handler(message, user)

        if TaalcBot.testers and not result:
            print(f"{handler.__name__} doesn't return any result")

        return result
    
    except Exception as ex:
        tb = traceback.format_tb(ex.__traceback__)
        tb = "\n".join(tb)
        err_msg = f'{ex}\n {tb}'
        if TaalcBot.error_prefix:
            err_msg = f'{TaalcBot.error_prefix}: {err_msg}'
        await message.reply(err_msg)                    
        raise ex


def msg_handler(*args):
    def handler_wrapper(handler):

        async def wrapper(message: types.Message, match: Match):
            # if message.chat.type in (ChatType.GROUP, ChatType.SUPERGROUP):
            result = await handle_with_tests(handler, message, match)
            return result
        
        for route in args:
            TaalcBot.msg_handlers[route] = wrapper
        return wrapper
    return handler_wrapper


def cmd_handler(*args):
    def handler_wrapper(handler):

        async def wrapper(message: types.Message):
            
            result = await handle_with_tests(handler, message)
            return result
        
        for route in args:
            TaalcBot.cmd_handlers[route] = wrapper
        return wrapper
    return handler_wrapper

async def _on_member_updated(event: ChatMemberUpdated, handler):
    try:
        user = TUser.user_by_tg_user(event.new_chat_member.user)                
        result =  await handler(event, user)
        
        return result
    except Exception as ex:
        tb = traceback.format_tb(ex.__traceback__)
        tb = "\n".join(tb)
        err_msg = f'{ex}\n {tb}'
        if TaalcBot.error_prefix:
            err_msg = f'{TaalcBot.error_prefix}: {err_msg}'
        await event.answer(err_msg)                    
        raise ex
    

def join_handler(handler):
    async def wrapper(event: ChatMemberUpdated):
        return await _on_member_updated(event, handler)
    TaalcBot.join_handlers.append(wrapper)
        
    return wrapper


def leave_handler(handler):
    async def wrapper(event: ChatMemberUpdated):
        return await _on_member_updated(event, handler)
    TaalcBot.leave_handlers.append(wrapper)
        
    return wrapper


def promoted_handler(handler):
    async def wrapper(event: ChatMemberUpdated):
        return await _on_member_updated(event, handler)
    TaalcBot.promoted_handlers.append(wrapper)
        
    return wrapper

def reaction_handler(handler):
    async def wrapper(reaction: types.MessageReactionUpdated):
        try:
            user = TUser.user_by_tg_user(reaction.user)
            result =  await handler(reaction, user)
            
            return result
        except Exception as ex:
            tb = traceback.format_tb(ex.__traceback__)
            tb = "\n".join(tb)
            err_msg = f'{ex}\n {tb}'
            if TaalcBot.error_prefix:
                err_msg = f'{TaalcBot.error_prefix}: {err_msg}'
             
            await reaction.bot.send_message(reaction.chat.id, err_msg)                    
            raise ex
    TaalcBot.reaction_handlers.append(wrapper)
        
    return wrapper


from .worker import Worker
from aiogram import Bot, Dispatcher, types, Router, filters, F
from aiogram.filters.command import Command, CommandStart
from aiogram.filters import ChatMemberUpdatedFilter, JOIN_TRANSITION, LEAVE_TRANSITION, PROMOTED_TRANSITION
from epure.resource.db.db import Db
import asyncio
from collections import OrderedDict
from typing import List
# from .teest.tester import Tester


class TaalcBot(Worker):
    bot_token:str
    bot:Bot
    dsp:Dispatcher
    config:object
    testers: list = []

    msg_handlers = OrderedDict()
    cmd_handlers = OrderedDict()

    join_handlers = []
    leave_handlers = []
    promoted_handlers = []
    reaction_handlers = []

    error_prefix = ''

    db: Db = None

    

    def __init__(self, bot_token: str, db:Db=None, config:object=None, error_prefix = ''):        

        self.bot_token = bot_token
        self.config = config
        self.db = db
        self.__class__.error_prefix = error_prefix


        self.bot = Bot(self.bot_token)        
        self.dsp = Dispatcher()
        
        for route, handler in reversed(self.msg_handlers.items()):
            self.dsp.message(F.text.regexp(route).as_("match"))(handler)

        for route, handler in self.cmd_handlers.items():
            self.dsp.message(Command(route))(handler)

        for handler in self.join_handlers:
            self.dsp.chat_member(ChatMemberUpdatedFilter(JOIN_TRANSITION))(handler)

        for handler in self.leave_handlers:
            self.dsp.chat_member(ChatMemberUpdatedFilter(LEAVE_TRANSITION))(handler)

        for handler in self.promoted_handlers:
            self.dsp.chat_member(ChatMemberUpdatedFilter(PROMOTED_TRANSITION))(handler)

        for handler in self.reaction_handlers:
            self.dsp.message_reaction()(handler)
   
 
    def start(self):
        try:
            asyncio.get_running_loop()
            asyncio.create_task(self._start())
        except RuntimeError as ex:
            asyncio.run(self._start())
        
    async def _start(self):
        await self.dsp.start_polling(self.bot, skip_updates=True)

    async def stop(self):
        try:
            asyncio.get_running_loop()
            asyncio.create_task(self._stop())
        except RuntimeError as ex:
            asyncio.run(self._stop())

    async def _stop(self):
        await self.dsp.stop_polling()

    # def create_parser() -> ArgumentParser:
    #     parser = ArgumentParser()
    #     parser.add_argument("--token", help="Telegram Bot API Token")
    #     parser.add_argument("--chat-id", type=int, help="Target chat id")
    #     parser.add_argument("--message", "-m", help="Message text to sent", default="Hello, World!")

    #     return parser



from ..finance.taalc_nft import TaalcNft
from .t_user import TUser
from .t_chat import TChat
from ..bidding.t_offer import TOffer
from epure import epure

@epure()
class TMessage():
    owner: TUser
    creator: TUser
    taalc_offer: TOffer
    taalc_nft: TaalcNft
    taalc_chat: TChat
    tg_id: int


from aiogram import Bot, Dispatcher
from aiogram.types import Message, Update
from ..taalc_bot import TaalcBot
import asyncio
from .bot_response import BotResponse
from .testing_message import TestingMessage
from ..tg_environment.t_user import TUser

class Tester:
    bot: Bot
    test_chat_id: int
    dsp: Dispatcher
    tested_bot: TaalcBot
    response: BotResponse
    msg_event: asyncio.Event
    waiting_delay: int
    _t_user: TUser = None

    @property
    def t_user(self):
        if not self._t_user:
            self._t_user = TUser.user_by_tg_user(self.bot)
        return self._t_user
    
    async def handler(self, message: Message):
        msg_user = message.from_user
        if msg_user.is_bot and message.chat.id == self.test_chat_id and \
              msg_user.id == self.tested_bot.bot.id and not self.msg_event.is_set():
            
            message = TestingMessage(
                text = message.text,
                message_id = message.message_id,
                date = message.date,
                chat = message.chat,
                message_thread_id = message.message_thread_id,
                from_user = message.from_user,
                via_bot = message.via_bot,
                reply_to_message = message.reply_to_message
            )
            self.response.response = message
            self.response.is_responded = True
            self.msg_event.set()

    async def reply(self, replied_msg: Message, msg_text: str, parse_mode: str=None) -> BotResponse:
        # msg_text = f'<blockquote>replied from {replied_msg.from_user.full_name}:\n{replied_msg.text}</blockquote>\n{msg_text}'
        sent_msg = await self.bot.send_message(self.test_chat_id, msg_text, parse_mode=parse_mode)
        sent_msg = TestingMessage(
            text = sent_msg.text,
            message_id = sent_msg.message_id,
            date = sent_msg.date,
            chat = sent_msg.chat,
            message_thread_id = sent_msg.message_thread_id,
            from_user = sent_msg.from_user,
            via_bot = sent_msg.via_bot,
            reply_to_message = replied_msg
        )
        
        res = await self._msg(sent_msg, parse_mode)
        return res

    async def msg(self, msg_text: str, parse_mode: str=None) -> BotResponse:
        sent_msg = await self.bot.send_message(self.test_chat_id, msg_text, parse_mode=parse_mode)
        sent_msg = TestingMessage(
            text = sent_msg.text,
            message_id = sent_msg.message_id,
            date = sent_msg.date,
            chat = sent_msg.chat,
            message_thread_id = sent_msg.message_thread_id,
            from_user = sent_msg.from_user,
            via_bot = sent_msg.via_bot
        )
        
        return await self._msg(sent_msg, parse_mode)

    async def _msg(self, sent_msg: Message, parse_mode: str=None) -> BotResponse:
        

        self.response = BotResponse(sent_msg)
        self.msg_event = asyncio.Event()        
        
        async def waiter():
            await asyncio.sleep(self.waiting_delay)
            self.msg_event.set()

        self.dsp.message()(self.handler)
        asyncio.create_task(self.dsp.start_polling(self.bot, skip_updates=True))
        asyncio.create_task(waiter())
        upd = Update(update_id=1, message=sent_msg)
        await self.tested_bot.dsp.feed_update(self.tested_bot.bot, upd)
        # asyncio.gather(self.dsp.start_polling(self.bot, skip_updates=True), self.tested_bot._start())

        
        await self.msg_event.wait()
        # await asyncio.sleep(3)
        await self.dsp.stop_polling()
        self.dsp.message.handlers.clear()

        return self.response
        

    def __init__(self, tester_bot_token: str, tested_bot: TaalcBot, test_chat_id: int,\
                 waiting_delay=3):
        
        self.bot = Bot(tester_bot_token)
        self.test_chat_id = test_chat_id
        self.dsp = Dispatcher()
        self.tested_bot = tested_bot
        self.waiting_delay = waiting_delay

        
        TaalcBot.testers.append(self)


from aiogram.types import Message, Update
from ..taalc_bot import TaalcBot


    
class TestingMessage(Message):
    # _testing_msg: Message

    async def _notify(self, message):
        if TaalcBot.testers:
            upd = Update(update_id=1, message=message)
            for tester in TaalcBot.testers:
                await tester.dsp.feed_update(tester.bot, upd)

    async def reply(self, text, *args, parse_mode=None, **kwargs):
        res = None
        try:
            res = await super().reply(*args, text, **kwargs, parse_mode=parse_mode)            
        except Exception as ex:
            text = f'<blockquote>replied from {self.from_user.full_name}:\n{self.text}</blockquote>\n{text}'
            res = await super().answer(*args, text, **kwargs, parse_mode='html')            
        finally:
            await self._notify(res)
            return res


    async def answer(self, text, *args, parse_mode=None, **kwargs):
        res = await super().answer(*args, text, **kwargs, parse_mode=parse_mode)
        await self._notify(res)
        return res


from aiogram.types import ChatMemberUpdated
from ..decorators import join_handler, leave_handler,\
      cmd_handler, promoted_handler, reaction_handler
from aiogram import Bot, Dispatcher, types, Router
from ..tg_environment.t_user import TUser

@join_handler
async def new_member(event: ChatMemberUpdated, user):
    await event.answer(f'На колени, животное, <b>{user}</b>!\n'+
                        'Прочитай наши правила, и потом не говори, что ты не знал, петух:\n'+
                        '<a href="https://t.me/polysap_rules/2">Правила полисап</a>', parse_mode="HTML")

@leave_handler
async def left_member(event: ChatMemberUpdated, user):
    await event.answer(f'{user} покинул нас, земля пухом')

@promoted_handler
async def promoted_member(event: ChatMemberUpdated, user):
    await event.answer(f'{user} получил новые звёздочки, респект')

@cmd_handler('start', 'stop')
async def process_cmd(message: types.Message, user: TUser): 
    await message.reply('Ты по русски говори')

@reaction_handler
async def process_reaction(reaction: types.MessageReactionUpdated, user):
    chat = reaction.chat
    await reaction.bot.send_message(chat.id, \
                f'{user} отреагировал на <a href="https://t.me/c/{chat.shifted_id}/{reaction.message_id}">сообщение</a>', parse_mode="HTML")


from .taalc_bot import TaalcBot
from epure.files import IniFile
import asyncio
import argparse
from epure.dbs import GresDb
from .handlers.gift import *
from .handlers.administration import *

class Config:
    pass

def get_bot(config):
    db = GresDb(config.db_conn_str,
        log_level=config.log_level, 
        default_namespace=config.default_namespace)
    db.connect()
    
    bot = TaalcBot(config.bot_token, db, config)    
    return bot

if __name__ == '__main__':
    # config = IniFile('./pyconfig.ini')    
    parser = argparse.ArgumentParser(description='bot token')
    parser.add_argument('--token', type=str, help='bot token', required=False)
    parser.add_argument('--config', type=str, help='configuration file', required=False)
    args = parser.parse_args()
    # config = {'bot_token': args.token}

    config = Config()
    if hasattr(args, 'token') and args.token:
        config.bot_token = args.token
    if hasattr(args, 'config') and args.config:
        config = IniFile(args.config)

    bot = get_bot(config)
    bot.start()


from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from epure import epure, proto
from .t_member import TMember
from aiogram.types.user import User
from ..finance.currency import Currency
if TYPE_CHECKING:
    from ..finance.currency_transaction import CurrencyTransaction
from aiogram.types import Message
# from epure.generics import Check

@epure()
class TUser(TMember):    
    first_name: str
    last_name: str
    username: str
    _wallet = None

    @property
    def wallet(self):
        if not self._wallet:
            from ..finance.wallet import Wallet
            self._wallet = Wallet(self)
        return self._wallet

    @classmethod
    def users(cls):
        res = TUser.resource.read()
        return res

    @classmethod
    def user_by_tg_user(cls, tg_user: User) -> TUser:
        res: TUser = None
        if not cls.tg_user_is_saved(tg_user):
            res = cls.save_user(tg_user)
        else:            
            res = cls.resource.read(telegram_id = tg_user.id)
            res = res[0]
        return res


    def __init__(self, user: User=None):
        # super().__init__()
        if user:
            self.telegram_id = user.id
            self.first_name = user.first_name
            self.last_name = user.last_name
            self.username = user.username

    @classmethod
    def tg_user_is_saved(cls, user: User) -> bool:
        users = cls.users()
        filtered = list(filter(lambda u: u.telegram_id == user.id, users))
        res = len(filtered) > 0
        return res
    
    @classmethod
    def save_user(cls, user: User):
        res = cls(user).save()
        return res
    
    
    def send_currency(self, to_user: TUser, currency: Currency, amount: float) -> CurrencyTransaction:
        from ..finance.currency_transaction import CurrencyTransaction
        res = CurrencyTransaction(self, to_user, currency, amount)
        res = res.save()

        return res
    
    def __str__(self):
        res = 'noname user'
        first_name = self.first_name if self.first_name else ''
        last_name = self.last_name if self.last_name else ''
        if self.username:
            res = f'@{self.username}'
        elif self.first_name or self.last_name:
            res = f'{first_name} {last_name}'
        return res


from .transaction_batch import TransactionBatch
from ..tg_environment.t_user import TUser
from datetime import datetime


class TaalcTransaction():
    transaction_batch: TransactionBatch
    sent_from: TUser
    sent_to: TUser
    transaction_time: datetime